#include <cstdio>
#include <cstring>
#include <cassert>
#include <inttypes.h>

using namespace std;

#include "cbp3_def.h"
#include "cbp3_framework.h"
#include "predictor.h"


// Constant value
static const int COMPNUM = 19;

static const int L[] = { ///< History length for Tagged Components
 5, 6, 8, 10, 12, 15, 19, 24, 30, 37,
 47, 59, 73, 92, 115, 144, 180, 225, 281 
};

static const int E[] = { ///< Number of entries for Tagged Components
 8, 8, 9, 9, 10, 10, 9, 9, 9, 9,
 9, 9, 9, 9, 8, 8, 8, 8, 8
};

static const int T[] = { ///< Tag width for Tagged Components
 13, 13, 14, 14, 15, 15, 16, 16, 17, 17,
 18, 18, 19, 19, 21, 21, 22, 22, 23
};

static const int H = L[COMPNUM-1]; ///< Longest history length

/// The BCTAGE class
class BCTAGE {
private:
  static const int B = 8; ///< Number of bits of index and tag of Base components
  static const int W = 5; ///< Number of ways(set-associative) of Base components

  static const int BBIT = 2; // Counter width for Base components
  static const int CBIT = 2; // Counter width for Tagged components
  static const int UBIT = 2; // Counter width for replacement information
  static const int PBIT = 8; // Policy seletion counter width
  static const int QBIT = 16; // Bit width for performance counter

private:
  //////////////////////////////////////////////
  // Constant values and Temporary values
  //////////////////////////////////////////////

  // Temporary value
  // (this variables are recalculate predict / train function)
  // Therefor, we do not count this as storage cost
  TaggedEntry<CBIT,UBIT> **gentry;

  //////////////////////////////////////////////
  // Storage for the predictor
  //////////////////////////////////////////////

  // Branch history updated in the retire stage...
  BITHIST cghr; // Global history
  BITHIST cphr; // Path history
  BITHIST cthr; // Target history
  uint32_t cpath; // Path hash value

  // Branch history updated in the fetch stage...
  BITHIST sghr; // Global history
  BITHIST sphr; // Path history
  BITHIST sthr; // Target history
  uint32_t spath; // Path hash value

  // Prediction tables
  BTaggedTable<BBIT,UBIT> base[W]; // BIM table
  GTaggedTable<CBIT,UBIT> *gtab; // Global pattern table

  // Other counters
  bool hard, easy; // 2-bit flags
  uint32_t randseed; // 16-bit randseed
  UCOUNTER<QBIT> mcount; // Performance counter
  UCOUNTER<QBIT> ecount; // Performance counter
  SCOUNTER<PBIT> P; // Confidence counter
  
  void print_budget() {
    int c = 0;
    int b = 0;

    // for branch history
    c += 32 * 2; // budgets for path hash(cpath, spath)
    c += cghr.budget();
    c += cphr.budget();
    c += cthr.budget();
    c += sghr.budget();
    c += sphr.budget();
    c += sthr.budget();
    printf("Global history: %d\n", c); b+=c;

    // for base predictor
    c = W * base[0].budget();
    printf("Base predictor: %d\n", c); b+=c;

    // for tagged component
    c = 0;
    for(int i=0; i<COMPNUM; i++) // tagged components
      c += gtab[i].budget();
    printf("Tagged components: %d\n", c); b+=c;

    // other counter / flags
    c = (2+2*QBIT+PBIT+16);
    printf("Other counters: %d\n", c); b+=c;

    // total budget count.
    printf("budget %d, verify:%d\n", b, b < (65*1024*8));
  }

protected:
  ///< random number generater
  // we implemented this to clarify the budget (16-bits)
  bool myrand() {
    randseed = (22695477 * randseed) + 1;
    randseed = randseed & ((1<<16)-1);
    return (randseed >> 15) & 1;
  }

  //////////////////////////////////////////////////////////
  // Prediction values
  //////////////////////////////////////////////////////////
  uint32_t getPrediction(PredictionResult<CBIT, UBIT> *result, uint32_t pc,uint32_t path,
                         const BITHIST& ghr, const BITHIST& phr, const BITHIST& thr) const {
    // Calcuates index for each tables
    // We can replace all gentry reference by following reference.
    // However, it increase the execution time significantly.
    // Therefore, we cache the address of gtable.
    GInputVector giv(pc, path, ghr, phr, thr);
    for(int i=0; i<COMPNUM; i++) {
      gentry[i] = &gtab[i].getEntry(giv);
    }

    bool NA = false; // Newly allocated entry bit
    // BIM cascading multiplexer (searching longest match entry)
    for(int i=0; i<COMPNUM; i++) {
      if(gtab[i].isBim()) {
        if(gtab[i].tagHit(giv)) {
          result->sbank = result->cbank;
          result->cbank = i;
          result->saddr = result->caddr;
          result->caddr = gentry[i]->adr;
          NA = false;
          result->lm_bim_entry = gentry[i]; // longest matched bim entry
        }
      }
    }

    // TAG cascading multiplexer (searching longest match entry)
    for(int i=0; i<COMPNUM; i++) {
      if(!gtab[i].isBim()) {
        if(gtab[i].getTag(giv) == gentry[i]->tag) {
          result->sbank = result->cbank;
          result->cbank = i;
          result->saddr = result->caddr;
          result->caddr = gentry[i]->adr;
          NA = ( (gentry[i]->repl == 0) && 
                 (gentry[i]->conf == 0) );
          result->lm_tag_entry = gentry[i]; //longest matched tag entry
        }
      }
    }

    // Select second prioritized one when the newly allocated entry.
    if(P < 0) {
      if(NA)
        return result->saddr;
    }
    return result->caddr;
  }

private:

public:
  BCTAGE(int comp, int minh, int maxh, int ebit, int tbit) :
    cghr(H),
    cphr(H),
    cthr(H),
    sghr(H),
    sphr(H),
    sthr(H),
    randseed(1)
  {
    // Allocate storage
    gentry = new TaggedEntry<CBIT,UBIT>*[COMPNUM];

    // Allocate storage
    for (int i = 0; i < W; ++i) {
      base[i].init(1 << B);
    }
    gtab = new GTaggedTable<CBIT,UBIT>[COMPNUM];
    for(int i=0; i<COMPNUM; i++) {
      gtab[i].init(E[i], T[i], L[i], i);
    }
    for (int i = 0; i < COMPNUM; ++i)
      if ((i & 3) == 1)
        gtab[i].setBim(true);

    // Allocate history
    cpath = 0;
    spath = 0;
  }

  // Initialize the predictor
  void init() {
    hard = false;
    easy = false;
    ecount = (ecount.max()/2);
    mcount = (mcount.max()/9);
    P = 0;

    print_budget();
  }

  ///////////////////////////////////////////////////////////////
  // Read / Update base predictor
  ///////////////////////////////////////////////////////////////

  // Read base predictor
  uint32_t read_base(uint32_t pc) {
    BaseInputVector biv(pc);
    for (int i = 0; i < W; ++i) {
      if (base[i].tagHit(biv))
        return base[i].getEntry(biv).adr;
    }
    // When missed in all ways, then it returns 0 as prediction.
    return 0;
  }

  // Update base table
  void update_base(uint32_t pc, uint32_t target, bool update, bool hit) {
    BaseInputVector biv(pc);
    // Update entry whose tag is hit
    for (int i = 0; i < W; ++i) {
      if (!base[i].tagHit(biv))
        continue;
      if(!update)
        return;
      TaggedEntry<CBIT, BBIT> &e = base[i].getEntry(biv);
      if(e.conf == 0)
        e.adr = target;
      if (hit)
        ++e.conf;
      else
        --e.conf;
      ++e.repl;
      return;
    }

    // allocate new entry
    if (update) {
      // Search low confidece entry for victim selection
      // When the multiple entry shows the lowest confidence,
      // Less referenced entry is selected as the victim.
      int u = 0;
      for (int i = 0; i < W; ++i) {
        TaggedEntry<CBIT, BBIT> &victim = base[u].getEntry(biv);
        TaggedEntry<CBIT, BBIT> &e = base[i].getEntry(biv);
        if ((e.conf < victim.conf) ||
            ((e.conf == victim.conf) && (e.repl < victim.repl)))
          u = i;
      }

      int v = base[u].getEntry(biv).repl;
      for(int i=0; i<W; i++) {
        if(v > 0) {
          int w = base[i].getEntry(biv).repl - v;
          if(w < 0)
            w = 0;
          base[i].getEntry(biv).repl = w;
        }
      }

      // Initialized newly allocated entry
      base[u].store(biv, target);
    }
  }

  /// Try to allocate an entry in BIM-mode tables
  /// @return allocated or not
  bool allocate_bim(const InputVector& giv, uint32_t target) {
    // Find a least useful BIM entry
    int r = -1;
    for(int i=0; i<COMPNUM; i++) {
      if(gtab[i].isBim()) {
        if(r < 0) r = i;

        if(gentry[i]->repl < gentry[r]->repl) {
          r = i;
        } else if(gentry[i]->repl == gentry[r]->repl) {
          if(gentry[i]->conf < gentry[r]->conf) {
            r = i;
          } else if(gentry[i]->conf == gentry[r]->conf) {
            if(myrand()) r = i;
          }
        }
      }
    }
    // If there's no BIM component, we cannot allocate
    if (r < 0)
      return false;

    // Update replacement information
    if(gentry[r]->repl > 0) {
      for(int i=0; i<COMPNUM; i++) {
        if(gtab[i].isBim()) {
          --(gentry[i]->repl);
        }
      }
    }
    
    gentry[r]->tag = gtab[r].getTag(giv);
    gentry[r]->adr = target;
    gentry[r]->conf = 0;
    gentry[r]->repl = 1;
    return true;
  }

  // Allocating new entry
  void allocate_tag(const PredictionResult<CBIT, UBIT>& result, const InputVector& giv, uint32_t target) {
    // We only assign TAG component which have longer history length than 
    // that of the longest matching components 
    int lmc = result.cbank + 1;
    if (gtab[result.cbank].isBim())
      lmc = 0;

    // Bank decision
    int allocbank = -1;

    // Decide minimum count & useful count of minimum value.
    int mincnt = 0;
    int usecnt = 0;
    int minval = 1<<UBIT;
    for(int i = lmc; i<COMPNUM; i++) {
      if(!gtab[i].isBim()) {
        if(minval > gentry[i]->repl) {
          minval = gentry[i]->repl;
          mincnt=0;
          usecnt=0;
        }
        if(minval == gentry[i]->repl) {
          mincnt+=1;
          usecnt+=(gentry[i]->conf > 0) ? 0 : 1;
        }
      }
    }

    // Decide allocate tagged components
    bool alloc;
    alloc = false;
    for(int i=lmc; i<COMPNUM; i++) {
      if(!gtab[i].isBim()) {
        if(gentry[i]->repl == minval) {
          alloc = alloc || (usecnt == 1) || myrand();
          if(!(gentry[i]->conf > 0)) {
            if(alloc && (allocbank < 0)) {
              allocbank = i;
            }
            usecnt--;
          }
        }
      }
    }
    
    // Update replace counter
    alloc = false;
    for(int i=lmc; i<COMPNUM; i++) {
      if(!gtab[i].isBim()) {
        if(gentry[i]->repl == minval) {
          alloc = alloc || (mincnt == 1) || myrand();
          if(alloc && (allocbank < 0)) {
            allocbank = i;
          }
          mincnt--;
        }
      }
    }

    // Update replace counter
    for(int i=lmc; i<COMPNUM; i++) {
      if(!gtab[i].isBim()) {
        if(minval > 0) {
          --(gentry[i]->repl);
        }
      }
    }

    // Allcate new entry
    if(allocbank >= 0) {
      gentry[allocbank]->tag = gtab[allocbank].getTag(giv);
      gentry[allocbank]->adr = target;
      gentry[allocbank]->conf = 0;
      gentry[allocbank]->repl = 0;
    }
  }

  uint32_t predict(uint32_t pc, bool cond, bool indir) {
    // Provides prediction result
    // This function is called in the fetch stage.
    if(!indir) { return 0; }

    uint32_t base_pred = read_base(pc);
    PredictionResult<CBIT, UBIT> result(base_pred);
    return getPrediction(&result, pc, spath, sghr, sphr, sthr);
  }

  // Training phase
  void train(uint32_t pc, bool cond, bool indir, bool taken, uint32_t target) {

    if(!indir)
      return;
    uint32_t base_pred = read_base(pc);
    PredictionResult<CBIT, UBIT> result(base_pred);
    uint32_t pred_adr = getPrediction(&result, pc, cpath, cghr, cphr, cthr);
    bool pred_correct = target == pred_adr;
    GInputVector giv(pc, cpath, cghr, cphr, cthr);

    // Update replacement bit and the policy selection counter
    bool hit_c = result.caddr == target;
    bool hit_s = result.saddr == target;
    if(hit_c != hit_s) {
      if ( (gentry[result.cbank]->repl == 0) && 
           (gentry[result.cbank]->conf == 0) ) {
        if (hit_c)
          ++P;
        else
          --P;
      }
      gentry[result.cbank]->repl += hit_c? 1: -1;
    }

    bool base_taghit = base_pred != 0; ///< there's tag-matched base entry
    bool bim_taghit = result.lm_bim_entry != NULL; ///< there's tag-matched bim entry
    bool tag_taghit = result.lm_tag_entry != NULL; ///< there's tag-matched tag entry
    bool base_correct = base_pred == target; ///< prediction of base component was correct
    bool bim_correct = (bim_taghit && (result.lm_bim_entry->adr == target)); ///< prediction of bim component was correct
    bool tag_correct = (tag_taghit && (result.lm_tag_entry->adr == target)); ///< prediction of tag component was correct

    // Update longest matched tag component entry if exist
    if (tag_taghit) {
      TaggedEntry<CBIT, UBIT>& lm_tag_entry = *result.lm_tag_entry;

      // Update target address
      if(!pred_correct && (lm_tag_entry.conf == 0))
        lm_tag_entry.adr = target;

      // Update confidence counter
      lm_tag_entry.conf += tag_correct? 1: -1;

    } else if (bim_taghit) { // Update longest matched bim entry
      TaggedEntry<CBIT, UBIT>& lm_bim_entry = *result.lm_bim_entry;

      // Update target address
      if (!pred_correct && (lm_bim_entry.conf == 0))
        lm_bim_entry.adr = target;

      // Update confidence counter
      lm_bim_entry.conf += bim_correct? 1: -1;
      ++lm_bim_entry.repl;
    }

    // calculate and adapt bim / tag tables configuration
    if (!pred_correct)
      adaptiveRehashing(base_taghit, bim_taghit);

    // When all entries make wrong prediction, update Base tables.
    update_base(pc, target, !bim_taghit && !tag_taghit, base_correct);

    // If bim components don't have corresponding entry, allocate one.
    bool bim_allocated = false;
    if (!bim_taghit)
      bim_allocated = allocate_bim(giv, target);

    // Assign new TAG components
    bool alloc_new_entry =
      (result.caddr != target) &&
      !pred_correct &&
      (!hard || !bim_allocated);


    // Allocating new entry
    // Note that we only assign TAG component which have longer history length than 
    // that of the longest matching components 
    if(alloc_new_entry)
      allocate_tag(result, giv, target);
  }

  void adaptiveRehashing(bool base_taghit, bool bim_taghit) {
    // Collect information for adaptive rehashing
    if(!base_taghit) {
      if(!bim_taghit) {
        // Need more BIM mode entry
        mcount -= 32;
        ecount -= 256; // This parameter is (also) changed since first submission
      } else {
        // This indicates good balance
        mcount -= 8;
        ecount -= 32;
      }
    } else {
      // Too much BIM mode entry
      ++mcount;
      ++ecount;
    }

    if(hard) { ecount = ecount.min(); }
    if(easy) { mcount = mcount.max(); }

    if(mcount.isSaturatedPositive()) {
      if(hard) {
        hard = false;
        for (int i = 0; i < COMPNUM; ++i)
          if ((i & 1) == 1)
            gtab[i].setBim(false);
        // printf("NotHardMode\n") ;
      }
    }
    if(mcount.isSaturatedNegative()) {
      if(!hard) {
        hard = true;
        for (int i = 0; i < COMPNUM; ++i)
          if ((i & 1) == 1)
            gtab[i].setBim(true);
        // printf("HardMode\n") ;
      }
    }

    if(ecount.isSaturatedPositive()) {
      if(!easy) {
        easy = true;
        for (int i = 0; i < COMPNUM; ++i)
          if ((i & 3) == 1)
            gtab[i].setBim(false);
        // printf("EasyMode\n") ;
      }
    } 
    if(ecount.isSaturatedNegative()) {
      if(easy) {
        easy = false;
        for (int i = 0; i < COMPNUM; ++i)
          if ((i & 3) == 1)
            gtab[i].setBim(true);
        // printf("NotEasyMode\n") ;
      }
    }
  }


  // history update in the fetch stage.
  void supdate(uint32_t pc, bool cond, bool indir, bool taken, uint32_t target) {
    spath = (target >> 2) + (spath << 2);

    sghr.append(taken);
    sphr.append((pc     & 4) != 0);
    sthr.append((target & 4) != 0);
  }

  // history update in the retire stage.
  void update(uint32_t pc, bool cond, bool indir, bool taken, uint32_t target) {
    cpath = (target >> 2) + (cpath << 2);

    cghr.append(taken);
    cphr.append((pc     & 4) != 0);
    cthr.append((target & 4) != 0);
  }
};



// predictor lists
BCTAGE *bctage;

// count number of runs
uint32_t runs;

void PredictorInit() {
  runs = 0;
  bctage = new BCTAGE(19,5,279,9,13);
}

void PredictorReset() {
  bctage->init();
}

void PredictorRunEnd() {
  runs++;
  rewind_marked = false;
}

void PredictorExit() {
  delete bctage;
}

void PredictorRunACycle() {
  // get info about what uops are processed at each pipeline stage
  const cbp3_cycle_activity_t *cycle_info = get_cycle_info();

  // make prediction at fetch stage
  for (int i = 0; i < cycle_info->num_fetch; i++) {
    uint32_t fe_ptr = cycle_info->fetch_q[i];
    const cbp3_uop_dynamic_t *uop = &fetch_entry(fe_ptr)->uop;

    // Decode instruction
    bool cond  = (uop->type & IS_BR_CONDITIONAL) != 0;
    bool indir = (uop->type & IS_BR_INDIRECT   ) != 0;
    bool taken = (uop->br_taken) || (!cond) ;
    uint32_t gpred = 0;

    // Check whether the target instruction is brach.
    if (uop_is_branch(uop->type)) {
      // Prediction Phase
      gpred = bctage->predict(uop->pc, cond, indir);
      bctage->supdate(uop->pc, cond, indir, taken, uop->br_target);
      if(indir) { assert(report_pred(fe_ptr, false, gpred)); }
    }
  }

  // Training and update phase...
  for (int i = 0; i < cycle_info->num_retire; i++) {
    uint32_t rob_ptr = cycle_info->retire_q[i];
    const cbp3_uop_dynamic_t *uop = &rob_entry(rob_ptr)->uop;

    // Decode instruction
    bool cond  = (uop->type & IS_BR_CONDITIONAL) != 0;
    bool indir = (uop->type & IS_BR_INDIRECT   ) != 0;
    bool taken = (uop->br_taken) || (!cond) ;
    if (uop_is_branch(uop->type)) {
      // Update phase
      bctage->train (uop->pc, cond, indir, taken, uop->br_target);
      bctage->update(uop->pc, cond, indir, taken, uop->br_target);
    }
  }
}
