#ifndef __PREDICTOR_H__
#define __PREDICTOR_H__

#include <math.h>
#include <cstdlib>

/////////////////////////////////////////////////////////////////
// COUNTER COMPONENTS
// SCOUNTER and UCOUNTER are implementation of simple counter
/////////////////////////////////////////////////////////////////

/// Abstract saturated counter class
/// Do not instanciate directly. Use SCOUNTER or UCOUNTER instead.
class COUNTER {
protected:
  int ctr; /// The counter
public:
  virtual int max() const = 0;
  virtual int min() const = 0;
  virtual uint32_t budget() const = 0;
  virtual ~COUNTER(){};

  operator int() const {
    return ctr;
  }

  bool isSaturatedPositive() const {
    return ctr == max();
  }

  bool isSaturatedNegative() const {
    return ctr == min();
  }

  COUNTER& operator++() {
    if (!isSaturatedPositive())
      ++ctr;
    return *this;
  }

  COUNTER& operator--() {
    if (!isSaturatedNegative())
      --ctr;
    return *this;
  }

  COUNTER& operator+=(int i) {
    ctr += i;
    if (ctr > max())
      ctr = max();
    else if (ctr < min())
      ctr = min();
    return *this;
  }

  COUNTER& operator-=(int i) {
    ctr -= i;
    if (ctr > max())
      ctr = max();
    else if (ctr < min())
      ctr = min();
    return *this;
  }

  COUNTER& operator=(int i) {
    assert(min() <= i);
    assert(i <= max());
    ctr = i;
    return *this;
  }
};

/// Signed saturated counter
template <int BITS>
class SCOUNTER: public COUNTER {
public:
  virtual int max() const {return (1 << (BITS - 1)) - 1;}
  virtual int min() const {return -(1 << (BITS - 1));}
  virtual uint32_t budget() const {return BITS;}
  SCOUNTER() {ctr = 0;}
  SCOUNTER(int i) {ctr = i;}
};

/// Unsigned saturated counter
template <int BITS>
class UCOUNTER: public COUNTER {
public:
  virtual int max() const {return (1 << BITS) - 1;}
  virtual int min() const {return 0;}
  virtual uint32_t budget() const {return BITS;}
  UCOUNTER() {ctr = 0;}
  UCOUNTER(int i) {ctr = i;}
};

// Left bit-wise rotation
static uint32_t rotatel_bits(uint32_t bits, int shift, int width) {
  return ((bits << shift) | (bits >> (width - shift))) & ((1 << width) - 1);
}

/// sequence of bits class to hold (global|path|target) histories
class BITHIST {
  bool *bits;
  int size;
public:
  int pos;
  BITHIST(int size)
    : size(size),
      pos(size - 1) {
    bits = new bool[size];
    for (int i = 0; i < size; ++i)
      bits[i] = 0;
  }

  uint32_t budget() const {
    return size;
  }

  bool at(int p) const {
    return bits[(pos + 1 + p) % size];
  }

  uint32_t fold(int len, int width) const {
    uint32_t r = 0;
    for (int i = 0; i < len; ++i) {
      r ^= bits[(pos + 1 + i) % size] << (i % width);
    }
    return r;
  }

  void append(bool a) {
    bits[pos--] = a;
    if (pos < 0)
      pos = size - 1;
  }
};

/// InputVector: holds information to calculate index and tag for a table
class InputVector {
public:
  virtual uint32_t createIndex(int id, int width, int len, int mode) const = 0;
  virtual uint32_t createTag(int id, int width, int len, int mode) const = 0;
  virtual ~InputVector(){};
};

/// BaseInputVector for Base tables
class BaseInputVector: public InputVector {
public:
  BaseInputVector(const uint32_t& pc)
    : pc(pc) {}

  virtual uint32_t createIndex(int /*id*/, int width, int /*len*/, int /*mode*/) const {
    return (pc ^ (pc >> width) ^ (pc >> (2*width))) & ((1<<width)-1);
  }

  virtual uint32_t createTag(int /*id*/, int width, int /*len*/, int /*mode*/) const {
    uint32_t p = pc;
    p ^= (p >> 16);
    p ^= (p >>  8);
    return p & ((1 << 8) - 1);
  }

private:
  const uint32_t& pc;
};

/// GInputVector for Tagged(BIM or Tag) components
class GInputVector: public InputVector {
public:
  GInputVector(const uint32_t& pc, const uint32_t& path, const BITHIST& ghr, const BITHIST& phr, const BITHIST& thr)
    : pc(pc), path(path), ghr(ghr), phr(phr), thr(thr) {}

  virtual uint32_t createIndex(int id, int width, int len, int bim) const {
    uint32_t p = pc;
    int e = width;
    int l = bim? 0: len;

    p ^= (p >> id);
    p ^= (path & ((1<<(2*l))-1));
    uint32_t val = (p ^ (p >> e) ^ (p >> (2*e))) & ((1<<e)-1);
    int pos = 2*l;

    if (!bim) {
      val ^= rotatel_bits(ghr.fold(l, e), pos % e, e); pos += l;
      val ^= rotatel_bits(phr.fold(l, e), pos % e, e); pos += l;
      val ^= rotatel_bits(thr.fold(l, e), pos % e, e); pos += l;
    }
    return val;
  }

  virtual uint32_t createTag(int id, int width, int len, int bim) const {
    uint32_t p = pc;
    int e = width;
    int l = bim? 0: len;

    p ^= (path & ((1<<(2*l))-1));
    uint32_t val = (p ^ (p >> e) ^ (p >> (2*e))) & ((1<<e)-1);
    int pos = 2*l+id;
    if (!bim) {
      val ^= rotatel_bits(thr.fold(l, e), pos % e, e); pos += l;
      val ^= rotatel_bits(ghr.fold(l, e), pos % e, e); pos += l;
      val ^= rotatel_bits(phr.fold(l, e), pos % e, e); pos += l;
    }
    return val;
  }
private:
  const uint32_t& pc;
  const uint32_t& path;
  const BITHIST& ghr;
  const BITHIST& phr;
  const BITHIST& thr;
};

/////////////////////////////////////////////////////////////////
// Entry of tagged component
/////////////////////////////////////////////////////////////////

template<int CBIT, int UBIT>
class TaggedEntry {
public:
  uint32_t adr; // Target address
  uint32_t tag; // Tag value
  UCOUNTER<CBIT> conf; // Confidence counter
  UCOUNTER<UBIT> repl; // Replacement counter

  void init() {
    adr = 0;
    tag = 0;
    conf = 0;
    repl = 0;
  }
  uint32_t budget() {
    // return budget witout tag, because this class don't know tag width
    return conf.budget() + repl.budget() + 32;
  }
};

// Abstract table
class TaggedTable {
};

template<int CBIT, int UBIT>
class BTaggedTable {
private:
  int size;
public:
  TaggedEntry<CBIT, UBIT> *ent;
  BTaggedTable() {ent = NULL;}
  void init(int n) {
    size = n;
    ent = new TaggedEntry<CBIT, UBIT>[n];
    for (int i = 0; i < n; ++i) {
      ent[i].init();
    }
  }

  bool tagHit(const InputVector& iv) {
    return getEntry(iv).tag == getTag(iv);
  }

  TaggedEntry<CBIT, UBIT> &getEntry(const InputVector& iv) {
    return ent[iv.createIndex(0, 8, 0, 0)];
  }

  uint32_t getTag(const InputVector& iv) {
    return iv.createTag(0, 8, 0, 0);
  }

  void store(const InputVector& iv, uint32_t addr) {
    TaggedEntry<CBIT, UBIT> &e = getEntry(iv);
    e.tag = getTag(iv);
    e.adr = addr;
    e.conf = 0;
    e.repl = 1;
  }

  uint32_t budget() const {
    return size * (ent[0].budget() + 8);
  }

  virtual ~BTaggedTable() {delete[] ent;}
};

template<int CBIT, int UBIT>
class GTaggedTable {
public:
  TaggedEntry<CBIT, UBIT> *ent;
  GTaggedTable() {ent = NULL;}
  ~GTaggedTable() {delete[] ent;}
  int index_width;
  int tag_width;
  int history_len;
  int nr_entry;
  int comp;
  bool bim;
  void init(int index_width, int tag_width, int history_len, int comp) {
    this->index_width = index_width;
    this->tag_width = tag_width;
    this->history_len = history_len;
    this->comp = comp;
    nr_entry = 1 << index_width;
    bim = false;
    ent = new TaggedEntry<CBIT, UBIT>[nr_entry];
    for (int i = 0; i < nr_entry; ++i) {
      ent[i].init();
    }
  }

  bool tagHit(const InputVector& iv) {
    return getEntry(iv).tag == getTag(iv);
  }

  TaggedEntry<CBIT, UBIT> &getEntry(const InputVector& iv) {
    return ent[iv.createIndex(comp, index_width, history_len, bim)];
  }

  uint32_t getTag(const InputVector& iv) {
    return iv.createTag(comp, tag_width, history_len, bim);
  }

  void setBim(bool b) {
    bim = b;
  }

  bool isBim() const {
    return bim;
  }

  uint32_t budget() const {
    return nr_entry * (ent[0].budget() + tag_width);
  }

private:

};

/////////////////////////////////////////////////////////////////
// Prediction Results.
/////////////////////////////////////////////////////////////////
template<int CBIT, int UBIT>
struct PredictionResult {
  uint32_t caddr;
  uint32_t saddr;
  int cbank;
  int sbank;

  TaggedEntry<CBIT, UBIT> *lm_tag_entry; ///< Longest matched Tag Entry
  TaggedEntry<CBIT, UBIT> *lm_bim_entry; ///< Longest matched BIM Entry
  PredictionResult(uint32_t base=0) {
    cbank = sbank = -1;
    caddr = saddr = base;
    lm_tag_entry = NULL;
    lm_bim_entry = NULL;
  }
};


#endif
